import pytest
import numpy as np
from pathlib import Path

from guts_base import GutsBase
from guts_base.prob import conditional_survival
from pymob.inference.scipy_backend import ScipyBackend


from guts_base import PymobSimulator
from mempy.model import RED_SD_aging, BufferGUTS_SD_aging, RED_IT_aging,BufferGUTS_IT_aging, RED_IT, RED_SD, RED_IT_DA, RED_SD_DA, BufferGUTS_IT_DA, BufferGUTS_SD_DA, BufferGUTS_SD_CA, BufferGUTS_IT_CA
from mempy.input_data import read_exposure_survival


from tests.conftest import (
    OPENGUTS_ESTIMATES,
)


def assert_convergence(sim: GutsBase, rtol=0.05, atol=0.1):
    # test if inferer converged on the true estmiates
    pymob_estimates = sim.inferer.idata.posterior.median(("chain", "draw"))
    pymob_estimates_array = pymob_estimates.to_array().sortby("variable")
    openguts_estimates = OPENGUTS_ESTIMATES[sim._model_class.__name__.lower()] # type: ignore
    

    if openguts_estimates is None:
        # this explicitly skips testing the results, since they are not available,
        # but does not fail the test.
        pytest.skip()

    openguts_estimates = openguts_estimates[list(pymob_estimates.data_vars)]
    openguts_estimates_array = openguts_estimates.to_array().sortby("variable")
    np.testing.assert_allclose(
        pymob_estimates_array, 
        openguts_estimates_array, 
        rtol=rtol, atol=atol
    )



# run tests with the Simulation fixtures
def test_setup(sim: GutsBase):
    """Tests the construction method"""
    assert True


def test_simulation(sim: GutsBase):
    """Tests if a forward simulation pass can be computed"""
    # sim.dispatch_constructor()
    evaluator = sim.dispatch()
    evaluator()
    evaluator.results

    assert True

@pytest.mark.slow
def test_likelihood(sim: GutsBase):
    # TODO: fix this test. As long as probabilities are not too extreme, scipy and 
    # numpyro return the same solutions. I need to investigate why this breaks down
    # at some point. Simple fixes are nans (numpyro returns zeros for true nans, and
    # nans in survivors are forward filled (maybe this is also part of the solution?))
    # It works for RED_SD and RED_IT possibly, because the default parameters work better
    # for the problem. I need to put the computations side by side and investigate 
    # parameters and check where things depart. 
    pytest.skip()

    # set up numpyro backend
    sim.config.inference_numpyro.gaussian_base_distribution = False
    sim.set_inferer("numpyro")

    # create a likelihood function from the probability model
    loglik, _ = sim.inferer.create_log_likelihood(check=False, gradients=False, seed=1, return_type="full", scaled=False) # type: ignore

    # calculate the probabilities of piror and observations with numpyro
    theta = sim.model_parameter_dict
    ll = loglik(theta=theta)
    log_likelihood_observations_numpyro = ll[2]["survival_obs"]
    # log_likelihood_prior_numpyro = xr.Dataset(data_vars=ll[1]).to_array().values


    # Set up the scipy backend
    # TODO: Eventually improve that API so it is more user friendly
    conditional_survival.eps = float(sim.observations.eps.values)
    ScipyBackend._distribution.distribution_map["conditional_survival"] = (conditional_survival, {})
    
    # set the initial number
    sim.config.error_model.survival =\
        "conditional_survival(p=survival,n=survivors_at_start[:,[0]])"

    # set up scipy backend    
    sim.set_inferer("scipy")
    
    # calculate the probabilities of piror and observations with scipy
    results = sim.inferer.inference_model(theta=theta, observations=sim.observations)
    log_likelihood_observations_scipy = results["observations_prob"]["survival"]

    # test if likelihood computations are comparable
    np.testing.assert_almost_equal(
        log_likelihood_observations_scipy, 
        log_likelihood_observations_numpyro
    )



    results["theta_prob"]

    # TODO Compare vincents' LP calculations


@pytest.mark.slow
def test_export_import_sim(sim: GutsBase):
    sim.export()
    sim_export_results = sim.evaluate()
    imported_sim = GutsBase.from_directory(directory=sim.output_path)

    # just try accessing the posterior object
    sim_import_results = imported_sim.evaluate()

    # assert that the results from both simulations are exactly identical
    diff = sim_export_results - sim_import_results
    np.testing.assert_array_equal(diff.sum().to_array().values, 0)



@pytest.mark.integration
@pytest.mark.batch1
@pytest.mark.parametrize("backend", ["numpyro"])
def test_inference(sim: GutsBase, backend):
    # test inference with default settings
    sim.set_inferer(backend)
    sim.prior_predictive_checks()
    sim.inferer.run()
    sim.export()
    sim.posterior_predictive_checks()

    # Test report generation with reduced extent
    # for ECx computation (this can take a long time)
    sim.config.guts_base.ecx_mode = "draws"
    sim.config.guts_base.ecx_draws = 3
    sim.config.guts_base.ecx_force_draws = True
    sim.config.guts_base.ecx_estimates_x = [0.5]
    sim.config.guts_base.ecx_estimates_times = [2]
    sim.config.report.debug_report = True
    sim.report()

    assert_convergence(sim, rtol=0.05, atol=0.1)


@pytest.mark.integration
@pytest.mark.batch2
def test_estimate_parameters_separate_control_mortality(sim: GutsBase):
    sim.config.report.debug_report = True

    sim.estimate_parameters(
        forward_interpolate_exposure_data=True,
        background_mortality="pre-fit"
    )
    assert_convergence(sim, rtol=0.1, atol=0.1)


@pytest.mark.integration
@pytest.mark.batch3
def test_estimate_parameters_separate_control_mortality_transform_nuts(sim: GutsBase):
    max_exposure = float(sim.observations.exposure.max().values)
    sim.config.report.debug_report = True

    from guts_base.sim.transformer import (
        ConcentrationAdditionTransform, DamageAdditionTransform, GutsTransform
    )
    
    if sim._model_class in (RED_IT_DA, RED_SD_DA, BufferGUTS_IT_DA, BufferGUTS_SD_DA):
        transform_class = DamageAdditionTransform
        # FIXME: posterior recalculation tests fail when these transforms are applied
        # the error is very small, but it should be ensured that it is not a systematic
        # problem
        pytest.skip()
    elif sim._model_class in (BufferGUTS_SD_CA, BufferGUTS_IT_CA,):
        transform_class = ConcentrationAdditionTransform
        # FIXME: posterior recalculation tests fail when these transforms are applied
        # the error is very small, but it should be ensured that it is not a systematic
        # problem
        pytest.skip()
    else:
        transform_class = GutsTransform

    sim.estimate_parameters(
        forward_interpolate_exposure_data=True,
        transform_scalings={"x_in_factor": max_exposure, "time_factor": 1.0},
        background_mortality="pre-fit",
        inference_numpyro_kernel="nuts",
        transform_class=transform_class,
        inference_numpyro_nuts_warmup=1000,
        inference_numpyro_draws=500,
        inference_numpyro_nuts_chains=1,
    )
    assert_convergence(sim, rtol=0.1, atol=0.1)

@pytest.mark.integration
@pytest.mark.batch4
def test_estimate_parameters_nuts(sim: GutsBase):
    sim.config.report.debug_report = True
    sim.estimate_parameters(
        forward_interpolate_exposure_data=True,
        inference_numpyro_kernel="nuts",
        inference_numpyro_nuts_warmup=1000,
        inference_numpyro_draws=500,
        inference_numpyro_nuts_chains=1,
    )
    assert_convergence(sim, rtol=0.1, atol=0.1)

def _compare_models(sim_it, sim_sd):

    hb_values = np.logspace(-4, 0, 10)     # 1e-4 bis 1
    elt_values = np.linspace(0.5, 50, 12)  # 0.5 bis 50

    rtol = sim_sd.config.jaxsolver.rtol * 20
    atol = sim_sd.config.jaxsolver.atol * 20

    for hb in hb_values:
        for elt in elt_values:

            params = {
                "kd": 0.0,
                "hb": hb,
                "elt": elt,
            }

            result_it = sim_it.evaluate(parameters=params)
            result_sd = sim_sd.evaluate(parameters=params)

            survival_it = result_it.survival
            survival_sd = result_sd.survival

            # -------------------------------
            # Structural checks
            # -------------------------------
            assert survival_it.dims == survival_sd.dims, \
                f"Dim mismatch for hb={hb}, elt={elt}"

            assert "time" in survival_it.coords
            assert "time" in survival_sd.coords

            # -------------------------------
            # Interpolation (use IT time since denser)
            # -------------------------------

            survival_it_subset = survival_it.sel(time=survival_sd.time)


            np.testing.assert_allclose(
                survival_it_subset.values,
                survival_sd.values,
                rtol=rtol,
                atol=atol,
                err_msg=f"Survival differs for hb={hb}, elt={elt}"
            )

            
            # -------------------------------
            # Cumulative hazard comparison
            # -------------------------------
            cumhaz_sd = result_sd.H
            cumhaz_it = result_it.H
            cumhaz_it_subset = cumhaz_it.sel(time=cumhaz_sd.time)

            np.testing.assert_allclose(
                cumhaz_it_subset.values,
                cumhaz_sd.values,
                rtol=rtol,
                atol=atol,
                err_msg=f"Cumulative hazard differs for hb={hb}, elt={elt}"
            )
            

@pytest.mark.integration
def test_aging_models_equivalent(tmp_path):

    # -------------------------------------------------------------
    # Load data
    # -------------------------------------------------------------

    data_path = Path(__file__).parent.parent / "data"

    exposure_funcs, survival_data, num_expos, exposure_data = \
        read_exposure_survival(
            data_path,
            "Boscalid_Simon-delso_format_pool.xlsx",
            survival_name="Survival",
            exposure_name="Exposure",
            visualize=False,
            with_info_dict=False,
            with_raw_exposure_data=True
        )

    # -------------------------------------------------------------
    # Initialize models
    # -------------------------------------------------------------
    sim_it = PymobSimulator.from_model_and_dataset(
        exposure_data=exposure_data,
        survival_data=survival_data,
        model=RED_IT_aging(),
        output_directory=tmp_path / "it_model"
    )

    sim_sd = PymobSimulator.from_model_and_dataset(
        exposure_data=exposure_data,
        survival_data=survival_data,
        model=RED_SD_aging(),
        output_directory=tmp_path / "sd_model"
    )

    sim_it_buf = PymobSimulator.from_model_and_dataset(
        exposure_data=exposure_data,
        survival_data=survival_data,
        model=BufferGUTS_IT_aging(),
        output_directory=tmp_path / "it_model"
    )

    sim_sd_buf = PymobSimulator.from_model_and_dataset(
        exposure_data=exposure_data,
        survival_data=survival_data,
        model=BufferGUTS_SD_aging(),
        output_directory=tmp_path / "sd_model"
    )

    sim_it.dispatch_constructor()
    sim_sd.dispatch_constructor()

    sim_it_buf.dispatch_constructor()
    sim_sd_buf.dispatch_constructor()

    _compare_models(sim_it, sim_sd)
    _compare_models(sim_it_buf, sim_sd_buf)






'''
This file contains functions for handling input data, i.e. reading from files and converting to 
datastructures usable by this implementation.

Classes
----------
None

Functions
----------
multi_interp_treatment_generator()
    Creates interpolating treatment functions from exposure data.
create_datasets()
    Converts input-data to a form structured around the individual treatments.
read_exposure_survival()
    Reads a xlsx file with the exposure and survival data and converts it into useable datastructures.
get_xc()
    Extracts every exposure-function for a given treatment.
split_control_treatment_dataset():
    Splits a dataset into control and treatment datasets.
plot()
    Visualizes the Exposure- and Survival-data as imported by the input_data.read_exposure_survival()-function.
'''
import os
import re
import warnings
from typing import Optional, Dict, Tuple
import pint
import pandas as pd
from scipy import interpolate
import matplotlib.pyplot as plt
import numpy as np
from mempy.utils import MempyError

ureg = pint.UnitRegistry()

def multi_interp_treatment_generator(treatments):
    '''
    Creates interpolating treatment functions from exposure data.
    
    Parameters
    ----------
    treatments: pandas.DataFrame
        DataFrame containing the exposure concentrations with the treatments as columns and the timepoints as rows.

    Returns
    ----------
    treatment_funcs: dict
        Dictionary with the treatment-names as keys and the corresponding functions as values. 
    '''
    treatment_funcs = {}
    for treat in treatments:
        t = treatments[treat].index
        conc = treatments[treat].values
        conc1 = conc[~np.isnan(conc)]
        t = t[~np.isnan(conc)]
        s = interpolate.InterpolatedUnivariateSpline(t, conc1, k=1)
        s.__name__ = f'treatment_{treat.lower()}'
        treatment_funcs[treat] = s
    return treatment_funcs


def create_datasets(exposure_funcs, survival_data, treatment_list=None):
    '''
    Converts input-data to a form structured around the individual treatments.

    Parameters
    ----------
    exposure_funcs: dictionary
        Dict of exposures and treatments containing functions representing the external concentrations.
    survival_data: pandas.Dataframe
        Pandas.DataFrame containing the survival for the treatments and timepoints.
    treatment_list: list, optional
        List of strings to name the treatments. If None, the headers of the survival_data-dataframe are used.

    Returns
    ----------
    datasets: list
        List of datasets containing tuples with all exposures and the corresponding survival per treatment.
        In the following Form:
        [
            (exposure_1_treatment_1, exposure_2_treatment_1, ... , exposure_n_treatment_1, survival_treatment_1),
            (exposure_1_treatment_2, exposure_2_treatment_2, ... , exposure_n_treatment_2, survival_treatment_2),
            ...
            (exposure_1_treatment_m, exposure_2_treatment_m, ... , exposure_n_treatment_m, survival_treatment_m)
        ]
    '''
    datasets = []
    if treatment_list == None:
        survival_columns = list(survival_data.keys())
    else:
        survival_columns = treatment_list
    for treatment in survival_columns:
        current_exposure = []
        for sub in exposure_funcs:
            current_exposure.append(exposure_funcs[sub][treatment])
        datasets.append((*current_exposure, survival_data[treatment].dropna()))
    return datasets
    

def read_exposure_survival(
    path, 
    file_name, 
    exposure_name="Exposure", 
    survival_name="Survival", 
    info_name="Info", 
    with_info_dict=False, 
    with_raw_exposure_data=False, 
    visualize=False,
    na_values=["NA", "N/A", "N.A.", "na", "n/a", "n.a."],
):
    
    """
    Reads a xlsx file with the exposure and survival data and converts it into useable datastructures.
    
    Parameters
    --------
    path : str
        Path to the file
    file_name : str
        File name of the xlsx file with the survival and exposure data
    exposure_name : str, default='Exposure'
        Name of the datasheet(s) containing the exposure data
    survival_name : str, default='Survival'
        Name of the datasheet containing the survival data
    info_name : str, default='Info'
        Name of the datasheet containing additional informations
    with_info_dict : bool, default=False
        If True the information datasheet is extracted and converted into a dictionary
    with_raw_exposure_data : bool, default=False
        If the raw exposure data should be returned
    visualize : bool, default=False
        If True the imported data is visualized in two plots
    
    Returns
    --------
    exposure_funcs : dict 
        Dictionairy containing the exposure functions
    survival_data : DataFrame
        Pandas DataFrame containing the treatments as columns and times as index
    num_expos : int
        Number of different expositions that the organisms are subject to. This will be one in the standard guts models but e.g. the number of substances
        used in the exposure when considering a guts-mixtures model.
    info_dict : dict, optional
        Contents of the info datasheet, only returned if 'with_info_dict'=True
    """
    file = pd.ExcelFile(os.path.join(path, file_name))
    exposure_sheets = [exp for exp in file.sheet_names if exposure_name in str(exp)]
    num_expos = len(exposure_sheets)
    if num_expos < 1:
        raise ValueError(f'No sheet named {exposure_name} in the Excel file')
    else:
        exposure_funcs = {}
        raw_exposure_data = {}
        for exposure_sheet in exposure_sheets:
            exposure_data = file.parse(sheet_name=exposure_sheet,index_col=0, na_values=na_values)
            raw_exposure_data.update({exposure_sheet: exposure_data})
            
            # maybe this needs to be done after collecting all sheets and removing nan treatments
            exposure_funcs[exposure_sheet] = multi_interp_treatment_generator(exposure_data)

    survival_data = file.parse(sheet_name=survival_name,index_col=0)

    assert_identical_column_names(
        exposure_data=raw_exposure_data, 
        survival_data=survival_data
    )

    test_for_increasing_survival(survival_data=survival_data)

    if visualize:
        plot(exposure_funcs, survival_data)
    
    return_data = [exposure_funcs, survival_data, num_expos]

    if with_info_dict:
        df_info = file.parse( header=None, sheet_name=info_name)
        info_array = df_info.to_numpy()
        info_dict = {info[0]:info[1:] for info in info_array}
        return_data += [info_dict]

    if with_raw_exposure_data:
        return_data += [raw_exposure_data]

    return tuple(return_data)


def test_for_increasing_survival(survival_data):
    increasing_ids = survival_data.diff().max() > 0

    has_increasing_survival = False
    messages = []
    for _id, is_increasing in increasing_ids.items():
        if is_increasing:
            has_increasing_survival = True
            increasing_time = survival_data.loc[:, _id].diff() > 0
            increasing_time = survival_data.index[increasing_time]

            change_list = []
            for time_at in increasing_time:
                loc_increase = survival_data.index.get_loc(time_at)
                time_before = survival_data.index[loc_increase - 1]
                time_after = survival_data.index[loc_increase + 1]

                value_at = survival_data.loc[time_at, _id]
                value_before = survival_data.loc[time_before, _id]
                value_after = survival_data.loc[time_after, _id]

                msg_detail = "{tbe}:{vbe}->{tat}:{vat}->{taf}:{vaf}".format(
                    tbe=time_before, vbe=value_before,
                    tat=time_at, vat=value_at, 
                    taf=time_after, vaf=value_after
                )
                change_list.append(msg_detail)

            details = [f"{d}" for v, d in zip(increasing_time.values, change_list)]
            msg = f"id={_id} has increasing survival at [{survival_data.index.name}:survival]={details}"
            messages.append(msg)
            print(msg)

    if has_increasing_survival:
        raise MempyError(
            "Survival data contain increasing values over time. " +
            "Survival data with increasing values (from e.g. independent populations " +
            "without repeated observations) need a standard binomial error distribution, " +
            "which is currently not implemented. " +
            "Are you sure you mean to have increasing survival in your dataset?"+
            "The following ids have increasing survival values: \n - " + 
            "\n - ".join(messages)
        )

def read_excel_file(
    path: str, 
    sheet_name_prefix: str, 
    convert_time_to: Optional[str] = None
) -> Tuple[Dict[str, pd.DataFrame], pint.Unit]: 
    """Reads a file and extracts the column name and time unit from the first column
    if needed converts the time unit to a different unit.

    Parameters
    ----------

    sheet_name_prefix : str
        The name of the sheet or a common prefix for multiple sheets

    convert_time_to : Optional[str]
        Unit to convert the time vector to. Works only if the column has a time unit given
        in parentheses. E.g time [d], or time (seconds), or time [hours]. Any common time
        denotion works (see the documentation of pint: https://pint.readthedocs.io/en/stable/)
    """
    file = pd.ExcelFile(path)
    sheets = [
        name for name in file.sheet_names 
        if name.startswith(sheet_name_prefix)
    ]

    _data = {}
    for sheet in sheets:
        df = file.parse(sheet_name=sheet)

        time_name, time_unit = extract_value_from_brackets(df.columns[0])
        if time_unit is not None:
            parsed_time_unit = ureg(time_unit)
        else:
            warnings.warn(
                "No time unit found in the first column. You're playing a dangerous game. " +
                "Better be explicit and indicate the time unit (e.g. 'time [d]' or 'time [seconds]')"
            )

        df = df.rename(columns={df.columns[0]: time_name})
        df = df.set_index(time_name)

        if convert_time_to is not None:
            converted_time_unit = parsed_time_unit.to(convert_time_to)
            df.index = df.index * converted_time_unit.magnitude
        else:
            converted_time_unit = parsed_time_unit

        _data.update({sheet: df})

    return _data, converted_time_unit.units

def extract_value_from_brackets(text):
    """
    Extracts the value inside square or normal brackets from a string.

    Args:
        text: The input string.

    Returns:
        The value inside the brackets, or None if no brackets are found.
    """
    match_square = re.search(r"\[(.*?)\]", text)  #The regex here is the key part
    match_normal = re.search(r"\((.*?)\)", text)  #The regex here is the key part
    if match_square:
        bracket = match_square.group(1)  # group(1) captures the content inside the parentheses
        # Remove bracket and content, then trim spaces
        _text = text.replace(match_square.group(0), "").replace(" ","")
        return _text, bracket
    elif match_normal:
        bracket = match_normal.group(1)  # group(1) captures the content inside the parentheses
        _text = text.replace(match_normal.group(0), "").replace(" ","")
        return _text, bracket
    else:
        return text.strip(), None

def assert_identical_column_names(exposure_data, survival_data):
    """This function compares only the names of the columns (treatments)
    It does not compare identity of the time column names
    """
    columns = {}
    for key, df in exposure_data.items():
        c = df.columns.values
        columns.update({key: c})

    c = survival_data.columns.values
    columns.update({"Survival": c})

    column_df = pd.DataFrame(columns)
    column_df['matching'] = column_df.eq(column_df.iloc[:, 0], axis=0).all(1)

    if all(column_df["matching"]):
        return
    else:
        raise MempyError(
            "Mismatch in column names detected between survival and exposure sheets.\n" +
            "All columns in the spreadsheets are listed below. Please fix the non-matching " +
            "column names (indicated as matching->False):"+
            "\n\n" +
            str(column_df)
        )

def get_xc(treatment, exposure_funcs):
    '''
    Extracts every exposure-function for a given treatment.

    Parameters
    ----------
    treatment: String
        Name of the treatment to be handled.
    exposure_funcs: dict
        Dictionary of the exposures and treamtents containing the exposure-functions.
    
    Returns
    ----------
    xc: Tuple
        Tuple containing all exposure-functions for the given treatment.
    '''
    xc_list = []
    for exposure in exposure_funcs:
        xc_list.append(exposure_funcs[exposure][treatment])
    xc = tuple(xc_list)
    return xc


def split_control_treatment_dataset(datasets):
    """
    Splits a dataset into control and treatment datasets

    Parameters
    ----------
    datasets: dict
        A dictionary of all datasets that should be categorized.

    Returns
    ----------
    control_datasets: dict
        A dictionary containing the control datasets
    treat_datasets: dict
        A dictionary containing all other datasets
    """
    control_datasets = []
    treat_datasets = []
    for dataset in datasets:
        if "control" in dataset[-1].name.lower().strip():
            control_datasets.append(dataset)
        else:
            treat_datasets.append(dataset)

    return control_datasets, treat_datasets 


def plot(exposure_funcs, survival_data):
    """
    Visualizes the Exposure- and Survival-data as imported by the input_data.read_exposure_survival()-function.

    Parameters
    ----------
    exposure_funcs: dict
        Dictionary of functions representing the time dependant external exposure to be visualized
    survival_data: pandas.DataFrame
        Table of survival-data to be visualized
    """
    fig, ax = plt.subplots()
    times = survival_data.index
    for col in survival_data.columns:
        ax.plot(times, survival_data[col], label = col)
    ax.legend()
    ax.set(xlabel='Time', ylabel='Survival', title='Observed survival per treatment over Time')
    plt.show()

    fig, ax = plt.subplots()
    for substance in exposure_funcs.keys():
        for treatment in exposure_funcs[substance].keys():
            times = np.linspace(exposure_funcs[substance][treatment].get_knots()[0],exposure_funcs[substance][treatment].get_knots()[-1], 1000)
            ax.plot(times, exposure_funcs[substance][treatment](times), label = treatment)
    ax.legend()
    ax.set(xlabel='Time', ylabel='Concentration', title='Exposure-Concentrations/Treatments over Time')
    plt.show()



class MempyError(Exception):
    """Exception raised for custom error scenarios.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


UNIT_REGISTRY = {
    "hb":   "1/{T}",
    "m":    "{X}",
    "kd":   "1/{T}",
    "b":    "1/{T}/{X}",
    "beta": "",
    "eta":  "1/{T}",
    "eps":  "{X}",
    "w":    "{X}/{X_i}",
}